import NDK, { getRelayListForUser, NDKList, NDKUser } from '@nostr-dev-kit/ndk'
import { kinds } from 'nostr-tools'
import { MuteLists } from '../types'
import { UserProfile } from '../types/user'
import { hexToNpub, log, LogType, npubToHex, timeout } from '../utils'

export enum UserRelaysType {
  Read = 'readRelayUrls',
  Write = 'writeRelayUrls',
  Both = 'bothRelayUrls'
}

/**
 * Singleton class to manage metadata operations using NDK.
 */
export class MetadataController {
  private static instance: MetadataController
  private ndk: NDK
  private usersMetadata = new Map<string, UserProfile>()
  public adminNpubs: string[]
  public adminRelays = new Set<string>()
  public reportingNpub: string

  private constructor() {
    this.ndk = new NDK({
      explicitRelayUrls: [
        'wss://user.kindpag.es',
        'wss://purplepag.es',
        'wss://relay.damus.io/',
        import.meta.env.VITE_APP_RELAY
      ]
    })

    this.ndk
      .connect()
      .then(() => {
        console.log('NDK connected')
      })
      .catch((err) => {
        console.log('error in ndk connection', err)
      })

    this.adminNpubs = import.meta.env.VITE_ADMIN_NPUBS.split(',')
    this.reportingNpub = import.meta.env.VITE_REPORTING_NPUB
  }

  private setAdminRelays = async () => {
    const promises = this.adminNpubs.map((npub) => {
      const hexKey = npubToHex(npub)
      if (!hexKey) return null

      return getRelayListForUser(hexKey, this.ndk)
        .then((ndkRelayList) => {
          if (ndkRelayList) {
            ndkRelayList.writeRelayUrls.forEach((url) =>
              this.adminRelays.add(url)
            )
          }
        })
        .catch((err) => {
          log(
            true,
            LogType.Error,
            `❌ Error occurred in getting the instance of NDKRelayList for npub: ${npub}`,
            err
          )
        })
    })

    await Promise.allSettled(promises)
  }

  /**
   * Provides the singleton instance of MetadataController.
   *
   * @returns The singleton instance of MetadataController.
   */
  public static async getInstance(): Promise<MetadataController> {
    if (!MetadataController.instance) {
      MetadataController.instance = new MetadataController()

      await MetadataController.instance.setAdminRelays()
    }
    return MetadataController.instance
  }

  /**
   * Finds metadata for a given pubkey.
   *
   * @param hexKey - The pubkey to search for metadata.
   * @returns A promise that resolves to the metadata event.
   */
  public findMetadata = async (pubkey: string): Promise<UserProfile> => {
    const npub = hexToNpub(pubkey)

    const cachedMetadata = this.usersMetadata.get(npub)
    if (cachedMetadata) {
      return cachedMetadata
    }

    const user = new NDKUser({ npub })
    user.ndk = this.ndk

    const userProfile = await user.fetchProfile()
    if (userProfile) {
      this.usersMetadata.set(npub, userProfile)
    }

    return userProfile
  }

  /**
   * Finds metadata for admin user.
   *
   * @returns A promise that resolves to the metadata event.
   */
  public findAdminMetadata = async (): Promise<UserProfile> => {
    return this.findMetadata(this.adminNpubs[0])
  }

  public findUserRelays = async (
    hexKey: string,
    userRelaysType: UserRelaysType = UserRelaysType.Both
  ): Promise<string[]> => {
    log(true, LogType.Info, `ℹ Finding user's relays`, hexKey, userRelaysType)

    const ndkRelayListPromise = getRelayListForUser(hexKey, this.ndk)

    // Use Promise.race to either get the NDKRelayList instance or handle the timeout
    return await Promise.race([
      ndkRelayListPromise,
      timeout() // Custom timeout function that rejects after a specified time
    ])
      .then((ndkRelayList) => {
        if (ndkRelayList) return ndkRelayList[userRelaysType]
        return [] // Return an empty array if ndkRelayList is undefined
      })
      .catch((err) => {
        log(true, LogType.Error, err)
        return [] // Return an empty array if an error occurs
      })
  }

  public getMuteLists = async (
    pubkey?: string
  ): Promise<{
    admin: MuteLists
    user: MuteLists
  }> => {
    const adminMutedAuthors = new Set<string>()
    const adminMutedPosts = new Set<string>()

    const adminHexKey = npubToHex(this.reportingNpub)

    if (adminHexKey) {
      const muteListEvent = await this.ndk.fetchEvent({
        kinds: [kinds.Mutelist],
        authors: [adminHexKey]
      })

      if (muteListEvent) {
        const list = NDKList.from(muteListEvent)

        list.items.forEach((item) => {
          if (item[0] === 'p') {
            adminMutedAuthors.add(item[1])
          } else if (item[0] === 'a') {
            adminMutedPosts.add(item[1])
          }
        })
      }
    }

    const userMutedAuthors = new Set<string>()
    const userMutedPosts = new Set<string>()

    if (pubkey) {
      const userHexKey = npubToHex(pubkey)

      if (userHexKey) {
        const muteListEvent = await this.ndk.fetchEvent({
          kinds: [kinds.Mutelist],
          authors: [userHexKey]
        })

        if (muteListEvent) {
          const list = NDKList.from(muteListEvent)

          list.items.forEach((item) => {
            if (item[0] === 'p') {
              userMutedAuthors.add(item[1])
            } else if (item[0] === 'a') {
              userMutedPosts.add(item[1])
            }
          })
        }
      }
    }

    return {
      admin: {
        authors: Array.from(adminMutedAuthors),
        replaceableEvents: Array.from(adminMutedPosts)
      },
      user: {
        authors: Array.from(userMutedAuthors),
        replaceableEvents: Array.from(userMutedPosts)
      }
    }
  }

  /**
   * Retrieves a list of NSFW (Not Safe For Work) posts that were not specified as NSFW by post author but marked as NSFW by admin.
   *
   * @returns {Promise<string[]>} - A promise that resolves to an array of NSFW post identifiers (e.g., URLs or IDs).
   */
  public getNSFWList = async (): Promise<string[]> => {
    // Initialize an array to store the NSFW post identifiers
    const nsfwPosts: string[] = []

    // Convert the  public key (npub) to a hexadecimal format
    const hexKey = npubToHex(this.reportingNpub)

    // If the conversion is successful and we have a hexKey
    if (hexKey) {
      // Fetch the event that contains the NSFW list
      const nsfwListEvent = await this.ndk.fetchEvent({
        kinds: [kinds.Curationsets],
        authors: [hexKey],
        '#d': ['nsfw']
      })

      if (nsfwListEvent) {
        // Convert the event data to an NDKList, which is a structured list format
        const list = NDKList.from(nsfwListEvent)

        // Iterate through the items in the list
        list.items.forEach((item) => {
          if (item[0] === 'a') {
            // Add the identifier of the NSFW post to the nsfwPosts array
            nsfwPosts.push(item[1])
          }
        })
      }
    }

    // Return the array of NSFW post identifiers
    return nsfwPosts
  }
}
